from __future__ import annotations

import collections
import functools
import itertools
import logging
import platform
import sys
from typing import Callable, Deque, Generator, Iterable, TypeVar

import jax
import toolz

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")


@toolz.curry
def requires_gpu(func: Callable[P, T], linux_only: bool = False) -> Callable[P, T]:
    """
    Fail if function is executing on host without access to GPU(s).
    Useful for early detecting container runtime misconfigurations.

    Parameters
    ----------
    func:
        Function, which needs hardware acceleration.
    linux_only:
        If set to true, will ignore check on non-linux hosts.


    Returns
    -------

    func:
        Function with the same signature as original one.

    """

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if linux_only and platform.system().lower() != "linux":
            logging.info(
                "Not running on linux, and linux_only==True, ignoring GPU strategy check."
            )
            return func(*args, **kwargs)

        devices = jax.devices()
        logging.info(f"JAX devices -> {devices}")
        if devices[0].device_kind != "gpu":
            raise RuntimeError("No GPU available.")
        return func(*args, **kwargs)

    return wrapper


def prefetch_to_device(
    iterator: Iterable[T], size: int = 2
) -> Generator[T, None, None]:
    """
    Parameters
    ----------
    iterator: Iterable[T]
        The input iterator to prefetch elements from.
    size: int, optional
        The number of elements to prefetch at a time. Defaults to 2.

    Returns
    -------
    Generator[T, None, None]
        A generator that yields the prefetched elements from the iterator.

    Raises
    ------
    ValueError
        If more than one GPU device is detected.

    Notes
    -----
    This method is used to prefetch elements from an iterator to a GPU device. It checks if the device is GPU and then
    enqueues *up to* `size` elements from the iterator to a deque. It uses JAX's `tree_map` and `device_put` functions to move
    the elements to the GPU device. The generator yields the prefetched elements one at a time.
    """
    queue: Deque[T] = collections.deque()
    devices = jax.devices()
    if devices[0].device_kind != "gpu":
        logging.error("Prefetch must be used only with GPU")
        for i in iterator:
            yield i

    if len(devices) > 1:
        raise ValueError(
            "Prefetch must be used only with single GPU, for multi-GPU support us flax.jax_utils.prefetch_to_device."
        )

    def enqueue(n: int) -> None:
        """Enqueues *up to* `n` elements from the iterator."""
        for data in itertools.islice(iterator, n):
            queue.append(
                jax.tree_util.tree_map(lambda xs: jax.device_put(xs, devices[0]), data)
            )

    enqueue(size)  # Fill up the buffer.
    while queue:
        yield queue.popleft()
        enqueue(1)



from __future__ import annotations

import functools
from importlib import util
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    NamedTuple,
    Protocol,
    TypeVar,
)

from absl import app, flags, logging

from absl_extra.notifier import BaseNotifier, LoggingNotifier

T = TypeVar("T", bound=Callable)
FLAGS = flags.FLAGS
flags.DEFINE_string("task", default="main", help="Name of the function to execute.")

if util.find_spec("pymongo"):
    from pymongo import MongoClient
    from pymongo.collection import Collection
else:
    Collection = type(None)
    logging.warning("pymongo not installed.")

if util.find_spec("ml_collections"):
    from ml_collections import ConfigDict, config_flags
else:
    logging.warning("ml_collections not installed")
    ConfigDict = None


if TYPE_CHECKING:
    from absl_extra.callbacks import CallbackFn


class MongoConfig(NamedTuple):
    uri: str
    db_name: str
    collection: str


class _ExceptionHandlerImpl(app.ExceptionHandler):
    def __init__(self, name: str, notifier: BaseNotifier):
        self.name = name
        self.notifier = notifier

    def handle(self, exception: Exception) -> None:
        self.notifier.notify_task_failed(self.name, exception)


class _TaskFn(Protocol):
    def __call__(
        self, *, config: ConfigDict = None, db: Collection = None, **kwargs
    ) -> None:
        ...


_TASK_STORE: Dict[str, Callable[[...], None]] = dict()  # type: ignore


class NonExistentTaskError(RuntimeError):
    def __init__(self, task: str):
        super().__init__(
            f"Unknown task {task}, registered are {list(_TASK_STORE.keys())}"
        )


def _make_task_func(
    func: _TaskFn,
    *,
    name: str,
    notifier: BaseNotifier | Callable[[], BaseNotifier],
    config_file: str | None,
    init_callbacks: List[CallbackFn],
    post_callbacks: List[CallbackFn],
    db: Collection | None,
) -> _TaskFn:
    _name = name

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        app.install_exception_handler(_ExceptionHandlerImpl(name, notifier))  # type: ignore
        if util.find_spec("ml_collections") and config_file is not None:
            config = config_flags.DEFINE_config_file("config", default=config_file)
            config = config.value
            kwargs["config"] = config
        else:
            config = None
        if db is not None:
            kwargs["db"] = db

        for hook in init_callbacks:
            hook(_name, notifier=notifier, config=config, db=db)

        func(*args, **kwargs)

        for hook in post_callbacks:
            hook(_name, notifier=notifier, config=config, db=db)

    return wrapper


def register_task(
    *,
    name: str = "main",
    notifier: BaseNotifier | Callable[[], BaseNotifier] | None = None,
    config_file: str | None = None,
    mongo_config: MongoConfig | Mapping[str, Any] | None = None,
    init_callbacks: List[CallbackFn] | None = None,
    post_callbacks: List[CallbackFn] | None = None,
) -> Callable[[_TaskFn], None]:
    """
    Parameters
    ----------
    name : str, optional
        The name of the task. Default is "main".
    notifier : BaseNotifier | Callable[[], BaseNotifier] | None, optional
        The notifier object or callable that returns a notifier object. Default is None.
    config_file : str | None, optional
        The path to the configuration file. Default is None.
    mongo_config : MongoConfig | Mapping[str, Any] | None, optional
        The configuration object for MongoDB or a mapping of configuration values. Default is None.
    init_callbacks : List[CallbackFn] | None, optional
        The list of callback functions to be executed during task initialization. Default is None.
    post_callbacks : List[CallbackFn] | None, optional
        The list of callback functions to be executed after the task completes. Default is None.

    Returns
    -------
    Callable[[_TaskFn], None]
        The decorator function that registers the task.
    """
    from absl_extra.callbacks import (
        log_absl_flags_callback,
        log_shutdown_callback,
        log_startup_callback,
        log_jax_devices,
        log_tensorflow_devices,
    )

    if isinstance(notifier, Callable):  # type: ignore
        notifier = notifier()  # type: ignore
    if notifier is None:
        notifier = LoggingNotifier()

    if util.find_spec("pymongo") and mongo_config is not None:
        if isinstance(mongo_config, Mapping):
            mongo_config = MongoConfig(**mongo_config)
        db: Collection[Mapping[str, ...]] = (
            MongoClient(mongo_config.uri)
            .get_database(mongo_config.db_name)
            .get_collection(mongo_config.collection)
        )
    else:
        db = None

    if init_callbacks is None:
        init_callbacks = [log_absl_flags_callback, log_startup_callback]
        if util.find_spec("tensorflow"):
            init_callbacks.append(log_tensorflow_devices)
        if util.find_spec("jax"):
            init_callbacks.append(log_jax_devices)

    if post_callbacks is None:
        post_callbacks = [log_shutdown_callback]

    def decorator(func: _TaskFn) -> None:
        _TASK_STORE[name] = functools.partial(  # type: ignore
            _make_task_func,
            name=name,
            notifier=notifier,
            init_callbacks=init_callbacks,
            post_callbacks=post_callbacks,
            db=db,
            config_file=config_file,
        )(func)

    return decorator


def run(argv: List[str] | None = None, **kwargs):
    """
    Parameters
    ----------
    argv:
        CLI args passed to absl.app.run
    kwargs:
        Kwargs passed to entrypoint function.

    Returns
    -------

    """

    def select_main(_):
        task_name = FLAGS.task
        if task_name not in _TASK_STORE:
            raise NonExistentTaskError(task_name)
        func = _TASK_STORE[task_name]
        func(**kwargs)

    app.run(select_main, argv=argv)



from __future__ import annotations

from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Protocol,
    Sized,
    Tuple,
    Type,
    TypeVar,
    no_type_check,
)

import clu.metrics
import clu.periodic_actions
from absl import logging
from flax import struct
from flax.core import frozen_dict
from flax.training import early_stopping, train_state
from jaxtyping import Array, Float, Int, Key, jaxtyped
from keras.utils.generic_utils import Progbar

from absl_extra.jax_utils import prefetch_to_device

T = TypeVar("T", contravariant=True)
TS = TypeVar("TS", bound=train_state.TrainState, contravariant=True)
M = TypeVar("M", bound=clu.metrics.Collection, contravariant=True)
DatasetFactory = Callable[[], Iterable[Tuple[T, Int[Array, "batch classes"]]]]  # noqa


@struct.dataclass
class NanSafeAverage(clu.metrics.Average):
    def compute(self) -> float:
        if self.count == 0:
            return 0
        return super().compute()


class ApplyFn(Protocol[T]):
    def __call__(
        self,
        params: Dict[str, frozen_dict.FrozenDict],
        x_batch: T,
        train: bool = False,
        rngs: Dict[str, Key] | None = None,
        **kwargs,
    ) -> Float[Array, "batch classes"]:  # noqa
        ...


class TrainingHook(Protocol[TS, M]):
    def __call__(
        self,
        step: int,
        *,
        training_state: TS,
        training_metrics: M,
        **kwargs,
    ) -> None:
        ...


class ValidationHook(Protocol[M]):
    def __call__(
        self,
        step: int,
        *,
        validation_metrics: M,
        **kwargs,
    ) -> None:
        ...


class UncheckedReportProgress(clu.periodic_actions.ReportProgress):
    def __call__(self, step: int, **kwargs) -> bool:
        return super().__call__(int(step))

    @no_type_check
    def _init_and_check(self, step: int, t: float):
        """Initializes and checks it was called at every step."""
        if self._previous_step is None:
            self._previous_step = step
            self._previous_time = t
            self._last_step = step
        else:
            self._last_step = step


class UncheckedPeriodicCallback(clu.periodic_actions.PeriodicCallback):
    def __call__(self, step: int, *args, **kwargs) -> bool:
        return super().__call__(int(step), *args, **kwargs)

    @no_type_check
    def _init_and_check(self, step: int, t: float):
        """Initializes and checks it was called at every step."""
        if self._previous_step is None:
            self._previous_step = step
            self._previous_time = t
            self._last_step = step
        else:
            self._last_step = step


def save_as_msgpack(
    params: frozen_dict.FrozenDict, save_path: str = "model.msgpack"
) -> None:
    """
    Parameters
    ----------
    params : frozen_dict.FrozenDict
        The frozen dictionary object that contains the parameters to be saved.
    save_path : str, optional
        The file path where the msgpack file will be saved. Default is "model.msgpack".

    Returns
    -------
    None
        This method does not return any value.
    """
    logging.info(f"Writing {save_path}")
    msgpack_bytes: bytes = frozen_dict.serialization.to_bytes(params)
    with open(save_path, "wb+") as file:
        file.write(msgpack_bytes)


def load_from_msgpack(
    params: frozen_dict.FrozenDict, save_path: str = "model.msgpack"
) -> frozen_dict.FrozenDict:
    """
    Load model parameters from a msgpack file.

    Parameters
    ----------
    params : frozen_dict.FrozenDict
        The original parameters of the model.
    save_path : str, optional
        The path to the msgpack file containing the serialized parameters.
        Default is "model.msgpack".

    Returns
    -------
    params : frozen_dict.FrozenDict
        The loaded parameters.

    """
    logging.info(f"Reading {save_path}")

    with open(save_path, "rb") as file:
        bytes_data = file.read()

    params = frozen_dict.serialization.from_bytes(params, bytes_data)

    return params


class InvalidEpochsNumberError(RuntimeError):
    def __init__(self, value: int):
        super().__init__(f"Epochs must be greater than 0, but found {value}")


class MissingSizeError(RuntimeError):
    def __init__(self, size_key: str = "size"):
        super().__init__(f"Must provide {size_key}, for not sized iterable.")


@jaxtyped
def train_on_single_device(
    *,
    training_state: TS,
    training_dataset_factory: DatasetFactory,
    validation_dataset_factory: DatasetFactory,
    metrics_container_type: Type[M],
    # fmt: off
    training_step_func: Callable[[TS, T, Int[Array, "batch classes"]], Tuple[TS, M]], # noqa
    validation_step_func: Callable[[TS, T, Int[Array, "batch classes"]], M], # noqa
    # fmt: on
    training_hooks: List[TrainingHook[TS, M]] | None = None,
    validation_hooks: List[ValidationHook[M]] | None = None,
    epochs: int = 1,
    prefetch_buffer_size: int = 2,
    verbose: bool = False,
    num_training_steps: int | None = None,
) -> Tuple[Tuple[Dict[str, float], Dict[str, float]], frozen_dict.FrozenDict]:
    """
    Parameters
    ----------
    training_state : TS
        The initial state of the training process.
    training_dataset_factory : DatasetFactory
        A factory function that returns the training dataset.
    validation_dataset_factory : DatasetFactory
        A factory function that returns the validation dataset.
    metrics_container_type : Type[M]
        The type of container to store the metrics.
    training_step_func : Callable[[TS, T, Int[Array, "batch classes"]], Tuple[TS, M]]
        A function that performs a single training step. It takes the training state, input data, and target data as inputs,
        and returns the updated training state and metrics.
    validation_step_func : Callable[[TS, T, Int[Array, "batch classes"]], M]
        A function that performs a single validation step. It takes the training state, input data, and target data as inputs,
        and returns the metrics.
    training_hooks : List[TrainingHook[TS, M]] | None, optional
        A list of training hooks to be executed after each training step. Defaults to None.
    validation_hooks : List[ValidationHook[M]] | None, optional
        A list of validation hooks to be executed after each validation step. Defaults to None.
    epochs : int, optional
        The number of training epochs. Defaults to 1.
    prefetch_buffer_size : int, optional
        The size of the prefetch buffer for loading data. Defaults to 2.
    verbose : bool, optional
        Whether to display verbose output during training. Defaults to False.
    num_training_steps : int | None, optional
        The total number of training steps. If None, it is calculated based on the size of the training dataset.
        Defaults to None.

    Returns
    -------
    Tuple[Tuple[Dict[str, float], Dict[str, float]], frozen_dict.FrozenDict]
        A tuple containing the training and validation metrics, and the final training state parameters.
    """
    if epochs <= 0:
        raise InvalidEpochsNumberError(epochs)

    if training_hooks is None:
        training_hooks = []
    if validation_hooks is None:
        validation_hooks = []

    if verbose:
        if num_training_steps is None:
            training_dataset = training_dataset_factory()
            if isinstance(training_dataset, Sized):
                num_training_steps = len(training_dataset)
            else:
                raise MissingSizeError("num_training_steps")

    for epoch in range(epochs):
        if verbose:
            logging.info(f"Epoch {epoch+1}/{epochs}...")

        training_dataset = training_dataset_factory()
        if prefetch_buffer_size != 0:
            training_dataset = prefetch_to_device(
                training_dataset, prefetch_buffer_size
            )

        if verbose:
            logging.debug(f"Training epoch {epoch+1}/{epochs}...")

        training_metrics = metrics_container_type.empty()

        if verbose:
            pbar = Progbar(num_training_steps)

        for x_batch, y_batch in training_dataset:
            training_state, metrics_i = training_step_func(
                training_state, x_batch, y_batch
            )
            training_metrics = training_metrics.merge(metrics_i)

            if verbose:
                pbar.update(
                    int(training_state.step),
                    [
                        (f"train_{k}", float(v))
                        for k, v in training_metrics.compute().items()
                    ],
                )

            for hook in training_hooks:
                hook(
                    training_state.step,
                    training_state=training_state,
                    training_metrics=training_metrics,
                )

        validation_dataset = validation_dataset_factory()
        if prefetch_buffer_size != 0:
            validation_dataset = prefetch_to_device(
                validation_dataset, prefetch_buffer_size
            )
        validation_metrics = metrics_container_type.empty()

        for x_batch, y_batch in validation_dataset:
            metrics_i = validation_step_func(training_state, x_batch, y_batch)
            validation_metrics = validation_metrics.merge(metrics_i)

        if verbose:
            pbar.update(
                int(training_state.step),
                [
                    (f"val_{k}", float(v))
                    for k, v in validation_metrics.compute().items()
                ],
            )

        for val_hook in validation_hooks:
            val_hook(training_state.step, validation_metrics=validation_metrics)

            if isinstance(val_hook, early_stopping.EarlyStopping):
                if val_hook.should_stop:
                    break

    return (
        training_metrics.compute(),  # noqa
        validation_metrics.compute(),  # noqa
    ), training_state.params



import enum
import inspect
from typing import Type
from types import MethodType

import flax.linen as nn
import toolz
from chex._src.variants import (
    Variant,
    check_variant_arguments,
    VariantsTestCaseGenerator,
    _variant_decorators,
    _valid_kwargs_keys,
)


class LiftedVariantType(enum.Enum):
    WITH_LIFTED_JIT = enum.auto()
    WITHOUT_LIFTED_JIT = enum.auto()

    def __str__(self) -> str:
        return "_" + self.name.lower()


@toolz.curry
def _variants_fn(
    test_object: MethodType, **which_variants
) -> VariantsTestCaseGenerator:
    """Implements `variants` and `all_variants`."""

    # Convert keys to enum entries.
    which_variants = {
        LiftedVariantType[name.upper()]: var  # type: ignore
        for name, var in which_variants.items()
    }
    if isinstance(test_object, VariantsTestCaseGenerator):
        # Merge variants for nested wrappers.
        test_object.add_variants(which_variants)
    else:
        test_object = VariantsTestCaseGenerator(test_object, which_variants)

    return test_object


@toolz.curry
def variants(
    test_method: MethodType,
    with_lifted_jit: bool = False,
    without_lifted_jit: bool = False,
) -> VariantsTestCaseGenerator:
    return _variants_fn(
        test_method,
        with_lifted_jit=with_lifted_jit,
        without_lifted_jit=without_lifted_jit,
    )


@toolz.curry
@check_variant_arguments
def _with_lifted_jit(target: Type[nn.Module], *args, **kwargs) -> Type[nn.Module]:
    return nn.jit(target, *args, **kwargs)


@toolz.curry
@check_variant_arguments
def _without_lifted_jit(target: Type[nn.Module], *args, **kwargs) -> Type[nn.Module]:
    return target


_variant_decorators.update(
    dict(
        {
            LiftedVariantType.WITH_LIFTED_JIT: _with_lifted_jit,
            LiftedVariantType.WITHOUT_LIFTED_JIT: _without_lifted_jit,
        }
    )
)

# Expose variant objects.
without_lifted_jit = Variant("without_lifted_jit", _without_lifted_jit)
with_lifted_jit = Variant("with_lifted_jit", _with_lifted_jit)
ALL_VARIANTS = (without_lifted_jit, with_lifted_jit)
# Collect valid argument names from all variant decorators.
for fn_ in _variant_decorators.values():
    original_fn = fn_.func.__wrapped__
    _valid_kwargs_keys.update(inspect.getfullargspec(original_fn).args)
